/*
 * This file is a component of thundr, a software library from 3wks.
 * Read more: http://www.3wks.com.au/thundr
 * Copyright (C) 2013 3wks, <thundr@3wks.com.au>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *         http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.threewks.thundr.googleapis.bigquery;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import com.google.api.client.googleapis.auth.oauth2.GoogleCredential;
import com.google.api.client.http.AbstractInputStreamContent;
import com.google.api.client.http.ByteArrayContent;
import com.google.api.client.http.javanet.NetHttpTransport;
import com.google.api.client.json.gson.GsonFactory;
import com.google.api.services.bigquery.Bigquery;
import com.google.api.services.bigquery.Bigquery.Jobs.Insert;
import com.google.api.services.bigquery.model.ErrorProto;
import com.google.api.services.bigquery.model.GetQueryResultsResponse;
import com.google.api.services.bigquery.model.Job;
import com.google.api.services.bigquery.model.JobConfiguration;
import com.google.api.services.bigquery.model.JobConfigurationLoad;
import com.google.api.services.bigquery.model.JobConfigurationQuery;
import com.google.api.services.bigquery.model.JobReference;
import com.google.api.services.bigquery.model.JobStatus;
import com.google.api.services.bigquery.model.TableCell;
import com.google.api.services.bigquery.model.TableFieldSchema;
import com.google.api.services.bigquery.model.TableReference;
import com.google.api.services.bigquery.model.TableRow;
import com.google.api.services.bigquery.model.TableSchema;
import com.threewks.thundr.exception.BaseException;
import com.threewks.thundr.googleapis.bigquery.BigQueryJobStatus.JobState;
import com.threewks.thundr.logger.Logger;

public class GoogleBigQueryServiceImpl implements GoogleBigQueryService {

	private static final String CSV_CONTENT_TYPE = "application/octet-stream";

	private Bigquery bigQuery;

	public GoogleBigQueryServiceImpl(GoogleCredential googleBigQueryCredential) {
		this.bigQuery = new Bigquery.Builder(new NetHttpTransport(), new GsonFactory(), googleBigQueryCredential).setApplicationName("thundr-google-apis").build();
	}

	@Override
	public Bigquery getBigQuery() {
		return bigQuery;
	}

	@Override
	public String queryAsync(String query, String projectId) throws IOException {
		Job job = new Job();
		JobConfiguration config = new JobConfiguration();
		JobConfigurationQuery queryConfig = new JobConfigurationQuery();
		config.setQuery(queryConfig);

		job.setConfiguration(config);
		queryConfig.setQuery(query);

		Insert insert = bigQuery.jobs().insert(projectId, job);
		insert.setProjectId(projectId);
		JobReference jobReference = insert.execute().getJobReference();

		Logger.debug("Submitted job %s to BigQuery", jobReference.getJobId());
		return jobReference.getJobId();
	}

	@Override
	public List<List<String>> getAsyncQueryResults(String jobId, String projectId, boolean includeColumnNames) throws IOException {
		Job job = bigQuery.jobs().get(projectId, jobId).execute();
		String state = job.getStatus().getState();
		if (state.equals("DONE")) {
			GetQueryResultsResponse response = bigQuery.jobs().getQueryResults(projectId, job.getJobReference().getJobId()).execute();
			TableSchema schema = response.getSchema();
			List<TableRow> rows = response.getRows();
			return buildQueryResults(schema, rows, includeColumnNames);
		}

		// job is not done yet, return null
		return null;
	}

	@Override
	public String executeLoad(String projectId, String datasetId, String tableId, String csvData) throws IOException {
		Logger.debug("Loading CSV data into %s[%s.%s]", projectId, datasetId, tableId);
		TableReference tableReference = new TableReference().setProjectId(projectId).setDatasetId(datasetId).setTableId(tableId);
		JobConfigurationLoad loadConfig = new JobConfigurationLoad().setDestinationTable(tableReference).setAllowQuotedNewlines(true);
		AbstractInputStreamContent content = new ByteArrayContent(CSV_CONTENT_TYPE, csvData.getBytes());
		return runJobWithContent(projectId, new JobConfiguration().setLoad(loadConfig), content);
	}

	@Override
	public String executeLoadFromCloudStorage(String projectId, String datasetId, String tableId, List<String> sourceUris) throws IOException {
		Logger.debug("Loading CSV data from %s into %s[%s.%s]", sourceUris.toString(), projectId, datasetId, tableId);
		TableReference tableReference = new TableReference().setProjectId(projectId).setDatasetId(datasetId).setTableId(tableId);
		JobConfigurationLoad loadConfig = new JobConfigurationLoad().setDestinationTable(tableReference).setAllowQuotedNewlines(true).setSourceUris(sourceUris);
		return runJob(projectId, new JobConfiguration().setLoad(loadConfig));
	}

	@Override
	public String createTableAndLoadFromGoogleCloudStorage(String projectId, String datasetId, String tableId, String schema, List<String> sourceUris) throws IOException {
		Logger.debug("Loading CSV data from %s into %s[%s.%s]", sourceUris.toString(), projectId, datasetId, tableId);
		TableReference tableReference = new TableReference().setProjectId(projectId).setDatasetId(datasetId).setTableId(tableId);
		JobConfigurationLoad loadConfig = new JobConfigurationLoad().setDestinationTable(tableReference).setAllowQuotedNewlines(true).setSchemaInline(schema).setSourceUris(sourceUris);
		return runJob(projectId, new JobConfiguration().setLoad(loadConfig));
	}

	@Override
	public BigQueryJobStatus getJobStatus(String jobId, String projectId) {
		try {
			Job job = bigQuery.jobs().get(projectId, jobId).execute();

			if (job == null) {
				return new BigQueryJobStatus(jobId, JobState.UNKNOWN, "Job does not exist");
			}

			JobStatus status = job.getStatus();
			if (status == null) {
				return new BigQueryJobStatus(jobId, JobState.UNKNOWN, "Cannot determine job status");
			}

			String state = status.getState();
			String errorMsg = null;
			if (state != null) {
				ErrorProto error = status.getErrorResult();
				if (error != null) {
					errorMsg = error.getMessage();
				}
			}
			return new BigQueryJobStatus(jobId, JobState.getState(state), errorMsg);

		} catch (Exception e) {
			throw new BaseException(e, "Error getting status of job %s for project %s", jobId, projectId);
		}
	}

	/**
	 * Submit a job to be run asynchronously and return the id of the job.
	 * 
	 * @param projectId the project to run the job in.
	 * @param jobConfiguration the job configuration.
	 * @return the id of the running job.
	 * @throws IOException if there is an error submitting the job.
	 */
	private String runJob(String projectId, JobConfiguration jobConfiguration) throws IOException {
		return runJobWithContent(projectId, jobConfiguration, null);
	}

	/**
	 * Submit a job to be run asynchronously with optional content and return the id of the job.
	 * 
	 * @param projectId the project to run the job in.
	 * @param jobConfiguration the job configuration.
	 * @param content optional content to run the job with.
	 * @return the id of the running job.
	 * @throws IOException if there is an error submitting the job.
	 */
	private String runJobWithContent(String projectId, JobConfiguration jobConfiguration, AbstractInputStreamContent content) throws IOException {
		Job job = new Job().setConfiguration(jobConfiguration);
		Job runningJob = content != null ? bigQuery.jobs().insert(projectId, job, content).execute() : bigQuery.jobs().insert(projectId, job).execute();
		return runningJob.getJobReference().getJobId();
	}

	/**
	 * Build a list of results.
	 * 
	 * @param schema the table schema for the results.
	 * @param rows the rows of results.
	 * @param includeColumnNames a boolean indicating whether the column names should be included in the results or not.
	 * @return the results as a list of list of strings.
	 */
	static List<List<String>> buildQueryResults(TableSchema schema, List<TableRow> rows, boolean includeColumnNames) {
		List<List<String>> results = new ArrayList<List<String>>();
		if (includeColumnNames) {
			results.add(getColumnNames(schema));
		}
		results.addAll(getResultData(rows));
		return results;
	}

	/**
	 * Get a list of the column names from a {@link TableSchema}.
	 * 
	 * @param schema the table schema to extract the column names from.
	 * @return a list of strings.
	 */
	static List<String> getColumnNames(TableSchema schema) {
		List<String> columnNames = new ArrayList<String>();
		List<TableFieldSchema> tableFieldSchema = schema.getFields();
		for (TableFieldSchema fieldSchema : tableFieldSchema) {
			columnNames.add(fieldSchema.getName());
		}
		return columnNames;
	}

	/**
	 * Extract the result data from a list of {@link TableRow} objects.
	 * 
	 * @param rows a list of table rows to extract the result data from.
	 * @return a list of rows (represented as lists of strings).
	 */
	static List<List<String>> getResultData(List<TableRow> rows) {
		List<List<String>> results = new ArrayList<List<String>>();
		if (rows != null) {
			for (TableRow tableRow : rows) {
				List<String> row = new ArrayList<String>();
				List<TableCell> fields = tableRow.getF();
				for (TableCell cell : fields) {
					row.add((String) cell.getV());
				}
				results.add(row);
			}
		}
		return results;
	}

}
